function compexp_upward(inode)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Computes parent's function coefficients using two children's
%   coefficients.
% 
% global variables.
%   NODES: tree structure
%
%   NOTE: We are using the 3rd basis function.
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global NODES

if NODES(5,inode) == 0, 
    % inode is a leaf node, compute the Fourier coefficients directly.
    leafnode(inode);   
    sprintf('leaf node, inode = %0.5f', inode)
else 
    % find children
    child1=NODES(5,inode); child2=NODES(6,inode);
    
    % find children's coefficients.
    compexp_upward(child1);
    compexp_upward(child2);
   
    % combine children's coefficients to get parent's coefficients.
    if NODES(3,child1)==1 && NODES(3,child2)==2
        % parent node is the root node.
        rootnode(inode,child1,child2);
    else
        % parent node is an interior node.
        internode(inode,child1,child2);
        sprintf('interior node, inode = %0.3f', inode)
    end
end

return
end

% help functions
function leafnode(inode)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute the coefficients for a leaf node.
%
% Note: at the leaf level, the integral becomes int_a^b 
%   c*e^(-alpha*x^2+2i(w1*phi1+w2*phi2)) dx.
%   where alpha=G(z_k,z_k), w1 and w2 are the two variables,
%   and phi1 and phi2 are the two basis functions.
%
% Input:
%   inode: the index of the leaf node.
%
% global variables.
%   NODES: tree structure
%   COEFF: the function coefficients for each node.
%   TRANSCoef: the parent-children interaction coefficients.
%   ABvec(1:2,ntotal): the integration bound for each variable. 
%   ZLoc(ntotal): the location of the z-variables.
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global NODES TRANSCoef COEFF ABvec ZLoc

% x-variable integration bound, z-variable bound, z_j, and W-variable info.
nzj=NODES(2,inode); % index for the location ZLoc.
xa=ABvec(1,nzj); xb=ABvec(2,nzj); % x-variable interval.
za=NODES(7,inode); zb=NODES(8,inode); zj=ZLoc(nzj); zc=(za+zb)/2; % z-variable interval.
nw1=TRANSCoef(9,inode);nw2=TRANSCoef(10,inode); 
lw1=TRANSCoef(7,inode); lw2=TRANSCoef(8,inode);

% construct the filter function using error function.
h=2/nw1; eps=1/14; x=-2:h:2-h/2; 
FilterFunW1=(erf((3/2+x)/eps)-erf((-3/2+x)/eps))/2;
h=2/nw2; x=-2:h:2-h/2; 
FilterFunW2=(erf((3/2+x)/eps)-erf((-3/2+x)/eps))/2;

% find alpha, phi1, phi2
if NODES(3,inode)==0, % only one interval, root node.
    alpha=exp(zj-1)*exp(1-zj)/2;
elseif NODES(3,inode)==1, % left boundary
    alpha=exp(zj-zb)*sinh(-zj+zb);   
elseif NODES(3,inode)==2, % right boundary
    alpha=exp(-zj+za)*sinh(zj-za);
else % interior point.
    alpha=sinh(zj-za)*sinh(-zj+zb)/sinh(zb-za); 
end
phi1=cosh(zj-zc); phi2=sinh(zj-zc)/(zb-za);   

% Note that alpha should be guranteed nonnegative.
% generate the function values at the w1,w2 nodes.
hw1=2*lw1/nw1; W1Loc=-2*lw1:hw1:2*lw1-hw1/2;
hw2=2*lw2/nw2; W2Loc=-2*lw2:hw2:2*lw2-hw2/2;

% funvalue2=zeros([2*nw1,2*nw2]);
% tic
% for kw1=1:2*nw1,
%     for kw2=1:2*nw2,
%         ytemp=(phi1*W1Loc(kw1)+phi2*W2Loc(kw2))/sqrt(alpha);
%         % funvalue2(kw1,kw2)=exp(-xtemp^2)*sqrt(pi)*...
%         %     (erfz(xb*sqrt(alpha)-1i*xtemp)...
%         %    -erfz(xa*sqrt(alpha)-1i*xtemp))/2/sqrt(alpha); 
%         % as xtemp may be large, to avoid 0*inf, we use the Faddeeva
%         % function.
%         atemp=xa*sqrt(alpha);btemp=xb*sqrt(alpha);
%         funvalue2(kw1,kw2)=sqrt(pi)*(exp(2*1i*atemp*ytemp-atemp^2)*...
%             fadf(ytemp+1i*atemp)-exp(2*1i*btemp*ytemp-btemp^2)*...
%             fadf(ytemp+1i*btemp))/2/sqrt(alpha); 
%         % no need to flip, the ordering is the same as matrix, from top to bottom, from left to right.
%     end
% end
% toc

% now the vectorized version.
mat1=W1Loc'*ones([1,2*nw2]);
mat2=ones([2*nw1,1])*W2Loc;
vec1=reshape(mat1,[4*nw1*nw2,1]);
vec2=reshape(mat2,[4*nw1*nw2,1]);
pvec=(phi1*vec1+phi2*vec2)/sqrt(alpha);
atemp=xa*sqrt(alpha);btemp=xb*sqrt(alpha);
funvec=sqrt(pi)*(exp(2*1i*atemp*pvec-atemp^2).*...
    fadf(pvec+1i*atemp)-exp(2*1i*btemp*pvec-btemp^2).*...
    fadf(pvec+1i*btemp))/2/sqrt(alpha); 
funvalue2=reshape(funvec,[2*nw1,2*nw2]);
% norm(funvalue3-funvalue2)



ffun2=funvalue2.*(FilterFunW1'*FilterFunW2);    % filter the original function.
COEFF(:,:,inode)=fftshift(fft2(fftshift(ffun2)))/(4*nw1*nw2);

return
end

function rootnode(inode,child1,child2)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute the answer for the root node.
% Input:
%   child1,child2: should be 2 and 3.
%
% global variables.
%   NODES: tree structure
%   COEFS: the function coefficients for each node point.
%   TRANSCoef(1:11,#ofNodes): the children-parent coefficient relations,the length
%        of the fundamental period, and the number of Fourier terms for each variable.
%        w1=c1*w1p+c2*w2p+c3*tnew;
%        w2=c4*w1p+c5*w2p+c6*tnew;
%           w1p,w2p are the parent's 2 variables. 
%        TRANSCoef(1:6,inode): store c1-c6. 
%        TRANSCoef(7:8,inode): the interval length of w1 and w2. Outside,
%           the integrand can be considered as 0.
%        TRANSCoef(9-11,inode): the number of Fourier Coefficients
%           for w1, w2 of current node, and t-variable used to decouple
%           parent node.
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global COEFF TRANSCoef

% In this case, c1,c2,c4,c5=0, we only have one t-variable.

ntnew=TRANSCoef(11,inode); % number of terms for the t-variable.
% Filter function for t-variable.
h=2/ntnew; eps=1/14; x=-2:h:2-h/2; 
FilterFunt=(erf((3/2+x)/eps)-erf((-3/2+x)/eps))/2; % filter function.
dt=14/ntnew; TLoc=-14:dt:14-dt/2; % location of the t-variables.

% compute the function value in physical domain for child 1.
TLoc1=TLoc*TRANSCoef(3,child1);
TLoc2=TLoc*TRANSCoef(6,child1);
phil=zeros(size(TLoc));
nw1=TRANSCoef(9,child1);nw2=TRANSCoef(10,child1); % number of terms in child 1's w-variable.
lw1=TRANSCoef(7,child1);lw2=TRANSCoef(8,child1);
for kw1=1:2*nw1,
    for kw2=1:2*nw2,
        TLoc3=(kw1-nw1-1)*TLoc1/(2*lw1)+(kw2-nw2-1)*TLoc2/(2*lw2);
        phil=phil+COEFF(kw1,kw2,child1)*exp(1i*pi*TLoc3);
    end
end

% compute the function values in physical domain for child 2.
phir=zeros(size(TLoc));
TLoc1=TLoc*TRANSCoef(3,child2);
TLoc2=TLoc*TRANSCoef(6,child2);
nw1=TRANSCoef(9,child2);nw2=TRANSCoef(10,child2); % number of terms in child 1's w-variable.
lw1=TRANSCoef(7,child2);lw2=TRANSCoef(8,child2);

for kw1=1:2*nw1,
    for kw2=1:2*nw2,
        TLoc3=(kw1-nw1-1)*TLoc1/(2*lw1)+(kw2-nw2-1)*TLoc2/(2*lw2);
        phir=phir+COEFF(kw1,kw2,child2)*exp(1i*pi*TLoc3);
    end
end

phinew=(phil.*phir).*(FilterFunt); % phil, phir: real function values.
coefnew=fftshift(fft(fftshift(phinew)))/2/ntnew; % unscaled real Fourier coefficients

% now integrate each term analytically and then find the final answer.
k=-ntnew:ntnew-1; weight=exp(-k.^2*pi^2/4/(14)^2);

% the final integral is given by the dot product of the weight and
% coefficients. The output is a number.
COEFF(1,1,1)=dot(coefnew,weight);

return
end

function internode(inode,child1,child2)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute the answer for an interior node (non-leaf and non-root).
%
% Input:
%   inode: current node.
%   child1,child2: index of the two children.
%
% global variables.
%   COEFF: the function coefficients for each node point.
%   TRANSCoef(1:11,#ofNodes): the children-parent coefficient relations,the length
%        of the fundamental period, and the number of Fourier terms for each variable.
%        w1=c1*w1p+c2*w2p+c3*tnew;
%        w2=c4*w1p+c5*w2p+c6*tnew;
%           w1p,w2p are the parent's 2 variables. 
%        TRANSCoef(1:6,inode): store c1-c6. 
%        TRANSCoef(7:8,inode): the interval length of w1 and w2. Outside,
%           the integrand can be considered as 0.
%        TRANSCoef(9-11,inode): the number of Fourier Coefficients
%           for w1, w2 of current node, and t-variable used to decouple
%           parent node.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global COEFF TRANSCoef 

mc1=TRANSCoef(9,child1); nc1=TRANSCoef(10,child1);
lw1c1=TRANSCoef(7,child1);lw2c1=TRANSCoef(8,child1);
mc2=TRANSCoef(9,child2); nc2=TRANSCoef(10,child2);
lw1c2=TRANSCoef(7,child2);lw2c2=TRANSCoef(8,child2);
mcp=TRANSCoef(9,inode); ncp=TRANSCoef(10,inode);
lw1p=TRANSCoef(7,inode);lw2p=TRANSCoef(8,inode);
ntnew=TRANSCoef(11,inode);

% Filter function.
% construct the filter function using error function.
h=2/mcp; eps=1/14; x=-2:h:2-h/2; % filter fun and location of W1P.
FilterFunW1=(erf((3/2+x)/eps)-erf((-3/2+x)/eps))/2; 
h=2*lw1p/mcp; W1Loc=-2*lw1p:h:2*lw1p-h/2;
h=2/ncp; x=-2:h:2-h/2; % filter fun and location of W2P.
FilterFunW2=(erf((3/2+x)/eps)-erf((-3/2+x)/eps))/2; 
h=2*lw2p/ncp;W2Loc=-2*lw2p:h:2*lw2p-h/2;
h=2/ntnew; x=-2:h:2-h/2; %filter fun and location of tnew variable.
FilterFunt=(erf((3/2+x)/eps)-erf((-3/2+x)/eps))/2; 
dt=14/ntnew;TLoc=-14:dt:14-dt/2;

% coeftemp is used to store the temporary physical/frequency domain
%   values for the parent node.
coeftemp=zeros(2*[mcp,ncp,ntnew]);

for nnt=-ntnew:ntnew-1,
    tnow=TLoc(nnt+ntnew+1); 
    
    % compute the function values in physical domain for child 1.
    mvec=-mc1:mc1-1;
    C3vec=exp(1i*pi*mvec'*TRANSCoef(3,child1)*tnow/(2*lw1c1));
    nvec=-nc1:nc1-1;
    C6vec=exp(1i*pi*nvec*TRANSCoef(6,child1)*tnow/(2*lw2c1));
    coeft1=COEFF(:,:,child1).*(C3vec*C6vec); % updated Fourier coefficients for the specific t-location.
    
    % compute the function values in physical domain for child 2.
    clear mvec nvec C3vec C6vec;
    mvec=-mc2:mc2-1;
    C3vec=exp(1i*pi*mvec'*TRANSCoef(3,child2)*tnow/(2*lw1c2));
    nvec=-nc2:nc2-1;
    C6vec=exp(1i*pi*nvec*TRANSCoef(6,child2)*tnow/(2*lw2c2));
    coeft2=COEFF(:,:,child2).*(C3vec*C6vec);
    
    % Evaluate the updated Fourier coefficients at the parent's W1 and W2
    % nodes points. In the future, this should be done using NuFFT.
    valt1=evaluate_Fourier(coeft1,mc1,nc1,lw1c1,lw2c1,...
        TRANSCoef(1:6,child1),mcp,ncp,W1Loc,W2Loc);
    valt2=evaluate_Fourier(coeft2,mc2,nc2,lw1c2,lw2c2,...
        TRANSCoef(1:6,child2),mcp,ncp,W1Loc,W2Loc);
    valtp=(valt1.*valt2).*(FilterFunW1'*FilterFunW2); % parent's filtered value.
    
    % compute parent's new W1,W2 Fourier coefficients at different t-values.
    coeftemp(:,:,nnt+ntnew+1)=fftshift(fft2(fftshift(valtp)))/(4*mcp*ncp);
end

% now integrate the t-variable.
% tic
for mx=1:2*mcp,
    for nx=1:2*ncp,
        ttemp(1:2*ntnew)=coeftemp(mx,nx,1:2*ntnew);
        tval=ttemp.*FilterFunt; % filter the t-varialbe function.
        tfourier=(fftshift(fft(fftshift(tval))))/(2*ntnew); % Fourier coefficients.
        % now integrate each term analytically and then find the final answer.
        k=-ntnew:ntnew-1; weight=exp(-k.^2*pi^2/4/(14)^2);
        COEFF(mx,nx,inode)=dot(weight,tfourier);
    end
end
% toc

return
end

function val=evaluate_Fourier(coef,mc1,nc1,lw1,lw2,cvalues,mcp,ncp,W1Loc,W2Loc)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% This function evaluates the Fourier expansion of size mc1 and nc1 (*2) at
%   parent's locations of size mcp,ncp.
%
% This is currently implemented directly, in the future, NuFFT will be
% applied to accelerate the computation.
%
% Input:
%   coef: Child's Fourier coefficient.
%   mc1,nc1: size of child's 2D expansion.
%   cvalues: the 6 values stating the relation between parent's and
%       children's variables.
%   mcp,ncp: size of the parent's 2D expansion.
%   W1Loc,W2Loc: the uniform evaluation points for each variable.
%
% Output:
%   val: parent nodes' function values at the uniform nodes.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% tic
% val=zeros(2*[mcp,ncp]);
% 
% % first implementation, using direct formulation, should be replaced by 
% %   nufft in the future.
% for mp=-mcp:mcp-1,
%   for np=-ncp:ncp-1,
%     for mc=-mc1:mc1-1,
%       for nc=-nc1:nc1-1,
%         val(mp+mcp+1,np+ncp+1)=val(mp+mcp+1,np+ncp+1)+...
%         coef(mc+mc1+1,nc+nc1+1)*exp(1i*pi*((mc*cvalues(1)/(2*lw1)+...
%         nc*cvalues(4)/(2*lw2))*W1Loc(mp+mcp+1)+(mc*cvalues(2)/(2*lw1)+...
%         nc*cvalues(5)/(2*lw2))*W2Loc(np+ncp+1)));
%       end
%     end
%   end
% end
% toc

% now the nufft version for the 2D case, in the future, we may have to 
%   consider W1Loc, W2Loc, and t at the same time.
mat1=W1Loc'*ones([1,2*ncp]);
mat2=ones([2*mcp,1])*W2Loc;
vec1=reshape(mat1,[4*mcp*ncp,1]);
vec2=reshape(mat2,[4*mcp*ncp,1]);
pvec1=pi*(cvalues(1)*vec1+cvalues(2)*vec2)/(2*lw1);
pvec2=pi*(cvalues(4)*vec1+cvalues(5)*vec2)/(2*lw2);

iflag=1;eps=1e-12;
% tic
valvec=nufft2d2(4*mcp*ncp,pvec1,pvec2,iflag,eps,2*mc1,2*nc1,coef);
val=reshape(valvec,[2*mcp,2*ncp]);
% norm(val1-val)
% toc

return
end
