function compute_integral(inode)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Computes parent's function coefficients using two children's
%   coefficients.
% 
% global variables.
%   NODES: tree structure
%   COEFS: the function coefficients for each node point.
%   ABvec(1:2,ntotal): the integration bound for each variable. 
%   FourierInfo: the number of terms (FourierInfo.num)
%       and the fundamental period (FourierInfo.LPeriod, function is 
%       periodic in [-Lperiod, Lperiod].
%   XLoc: The evenly distributed sample points in [-Lperiod, Lperiod].
%   FilterFun(1:FourierInfo.num*2, 1:FourierInfo.num*2): the filter
%       function values, so the function is well-approximated in 
%       [-LPeriod/2,LPeriod/2], and decays to 0 smoothly outside.
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global NODES

if NODES(5,inode) == 0, 
    % inode is a leaf node, compute the Fourier coefficients directly.
    leafnode(inode);   
else 
    % find children
    child1=NODES(5,inode); child2=NODES(6,inode);
    
    % find children's coefficients.
    compute_integral(child1);
    compute_integral(child2);
   
    % combine children's coefficients to get parent's coefficients.
    if NODES(3,child1)==1 && NODES(3,child2)==2
        % parent node is the root node.
        bdrylr(child1,child2);
    elseif NODES(3,child1)==1 && NODES(3,child2)==4
        % parent node is a left boundary node.
        leftbdry(child1,child2);
    elseif NODES(3,child1)==3 && NODES(3,child2)==2
        % parent node is a right boundary node.
        rightbdry(child1,child2);
    else
        % parent node is an interior node.
        interiornode(child1,child2);
    end
end

return
end

% help functions
function leafnode(inode)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute the coefficients for a leaf node.
% Input:
%   inode: the index of the leaf node.
%
% global variables.
%   NODES: tree structure
%   COEFS: the function coefficients for each node point.
%   ABvec(1:2,ntotal): the integration bound for each variable. 
%   FourierInfo: the number of terms (FourierInfo.num)
%       and the fundamental period (FourierInfo.LPeriod, function is 
%       periodic in [-Lperiod, Lperiod].
%   XLoc: The evenly distributed sample points in [-Lperiod, Lperiod].
%   FilterFun(1:FourierInfo.num*2, 1:FourierInfo.num*2): the filter
%       function values, so the function is well-approximated in 
%       [-LPeriod/2,LPeriod/2], and decays to 0 smoothly outside.
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global NODES COEFF ABvec XLoc FilterFun

a=ABvec(1,NODES(2,inode)); b=ABvec(2,NODES(2,inode));
if NODES(3,inode)==1, % left boundary, one variable function.
    % analytical solution is 1/2 e^(-t^2) Sqrt(Pi)(erf(b-it)-erf(a-it)), 
    % we have a single variable function.
    funvalue=1/2*sqrt(pi)*exp(-XLoc.^2).*(erfz(b-1i*XLoc)-erfz(a-1i*XLoc));
    % funvalue2=1/2*sqrt(pi)*(fadf(XLoc+1i*a).*exp(-a^2+2*1i*a*XLoc)-fadf(XLoc+1i*b).*exp(-b^2+2*1i*b*XLoc));
    % filtered function value.
    ffun=funvalue.*FilterFun;
    % find the Fourier coefficients.
    COEFF(:,1,inode)=(fftshift(fft(fftshift(ffun))))';
elseif NODES(3,inode)==2, % right boundary, one variable function.
   % the analytical solution is 1/2 e^(-t^2) Sqrt(Pi)(erf(b+it)-erf(a+it))
    funvalue=1/2*sqrt(pi)*exp(-XLoc.^2).*(erfz(b+1i*XLoc)-erfz(a+1i*XLoc));
    
    % filtered function value.
    ffun=funvalue.*FilterFun;
    % find the Fourier coefficients.
    COEFF(:,1,inode)=(fftshift(fft(fftshift(ffun))))'; 
else
    % analytical function is i/2*(e^(2ia(t2-t1))-e^(2ib(t2-t1))/(t2-t1)
    x1mx2=XLoc-XLoc(1)*ones(size(XLoc));
    cvector=1i/2*(-exp(-2*1i*b*x1mx2)+exp(-2*1i*a*x1mx2))./(-x1mx2);
    cvector(1)=b-a;
    rvector=1i/2*(-exp(2*1i*b*x1mx2)+exp(2*1i*a*x1mx2))./x1mx2;
    rvector(1)=(b-a);
    
    funvalue2=toeplitz(cvector,rvector); % no need to flip, the ordering is the same as matrix, from top to bottom, from left to right.
    
    ffun2=funvalue2.*(FilterFun'*FilterFun);    % filter the original function.
    COEFF(:,:,inode)=fftshift(fft2(fftshift(ffun2)));
end
   
end

function bdrylr(child1,child2)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute the answer for the root node.
% Input:
%   child1,child2: should be 2 and 3.
%
% global variables.
%   NODES: tree structure
%   COEFS: the function coefficients for each node point.
%   ABvec(1:2,ntotal): the integration bound for each variable. 
%   FourierInfo: the number of terms (FourierInfo.num)
%       and the fundamental period (FourierInfo.LPeriod, function is 
%       periodic in [-Lperiod, Lperiod].
%   XLoc: The evenly distributed sample points in [-Lperiod, Lperiod].
%   FilterFun(1:FourierInfo.num*2, 1:FourierInfo.num*2): the filter
%       function values, so the function is well-approximated in 
%       [-LPeriod/2,LPeriod/2], and decays to 0 smoothly outside.
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global COEFF FourierInfo 

% compute the function values in physical domain.
phil=fftshift(ifft(fftshift(COEFF(:,1,child1))));
phir=fftshift(ifft(fftshift(COEFF(:,1,child2))));
phinew=phil.*phir; % phil, phir: real function values.
coefnew=fftshift(fft(fftshift(phinew)))/2/FourierInfo.num; % unscaled real Fourier coefficients

% now integrate each term analytically and then find the final answer.
k=-FourierInfo.num:FourierInfo.num-1;
weight=exp(-k.^2*pi^2/4/(FourierInfo.LPeriod)^2);

% the final integral is given by the dot product of the weight and
% coefficients. The output is a number.
COEFF(1,1,1)=dot(coefnew,weight);

end

function leftbdry(child1,child2)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute the answer for the left boundary.
% Input:
%   child1,child2: index of the two children.
%
% global variables.
%   NODES: tree structure
%   COEFS: the function coefficients for each node point.
%   ABvec(1:2,ntotal): the integration bound for each variable. 
%   FourierInfo: the number of terms (FourierInfo.num)
%       and the fundamental period (FourierInfo.LPeriod, function is 
%       periodic in [-Lperiod, Lperiod].
%   XLoc: The evenly distributed sample points in [-Lperiod, Lperiod].
%   FilterFun(1:FourierInfo.num*2, 1:FourierInfo.num*2): the filter
%       function values, so the function is well-approximated in 
%       [-LPeriod/2,LPeriod/2], and decays to 0 smoothly outside.
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global NODES COEFF FourierInfo 

phil=fftshift(ifft(fftshift(COEFF(:,1,child1)))); % phil: real function values.
phir=fftshift(ifft(fftshift(COEFF(:,:,child2))))/2/FourierInfo.num; 
%above: ifft is applied to first variable. So phir is rescaled to give the real function values.
phinew=phir.*(phil*ones([1,FourierInfo.num*2])); % real function values.
coefnew=fftshift(fft(fftshift(phinew)));

% now integrate each term analytically and then find the final answer.
k=-FourierInfo.num:FourierInfo.num-1;
weight=exp(-k.^2*pi^2/4/(FourierInfo.LPeriod)^2);

% the final integral is given by the product of the weight and
% coefficients. The output is a vector.
iparent=NODES(4,child1);
COEFF(:,1,iparent)=(weight*coefnew)'; % as coefnew is scaled, COEFF is also scaled.

end

function rightbdry(child1,child2)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute the answer for the right boundary.
% Input:
%   child1,child2: index of the two children.
%
% global variables.
%   NODES: tree structure
%   COEFS: the function coefficients for each node point.
%   ABvec(1:2,ntotal): the integration bound for each variable. 
%   FourierInfo: the number of terms (FourierInfo.num)
%       and the fundamental period (FourierInfo.LPeriod, function is 
%       periodic in [-Lperiod, Lperiod].
%   XLoc: The evenly distributed sample points in [-Lperiod, Lperiod].
%   FilterFun(1:FourierInfo.num*2, 1:FourierInfo.num*2): the filter
%       function values, so the function is well-approximated in 
%       [-LPeriod/2,LPeriod/2], and decays to 0 smoothly outside.
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global NODES COEFF FourierInfo 
temp=COEFF(:,:,child1);
phil=fftshift(ifft(fftshift(temp')))/2/FourierInfo.num; 
% above: ifft applies to the 2nd variable, so transpose. rescale to get the
% real function values.
phir=fftshift(ifft(fftshift(COEFF(:,1,child2)))); % phir the real function values.
phinew=phil.*(phir*ones([1,FourierInfo.num*2])); 
coefnew=fftshift(fft(fftshift(phinew)));

% now integrate each term analytically and then find the final answer.
k=-FourierInfo.num:FourierInfo.num-1;
weight=exp(-k.^2*pi^2/4/(FourierInfo.LPeriod)^2);

% the final integral is given by the product of the weight and
% coefficients. The output is a vector.
iparent=NODES(4,child1);
COEFF(:,1,iparent)=(weight*coefnew)';

end

function interiornode(child1,child2)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute the answer for a interior node.
% Input:
%   child1,child2: index of the two children.
%
% global variables.
%   NODES: tree structure
%   COEFS: the function coefficients for each node point.
%   ABvec(1:2,ntotal): the integration bound for each variable. 
%   FourierInfo: the number of terms (FourierInfo.num)
%       and the fundamental period (FourierInfo.LPeriod, function is 
%       periodic in [-Lperiod, Lperiod].
%   XLoc: The evenly distributed sample points in [-Lperiod, Lperiod].
%   FilterFun(1:FourierInfo.num*2, 1:FourierInfo.num*2): the filter
%       function values, so the function is well-approximated in 
%       [-LPeriod/2,LPeriod/2], and decays to 0 smoothly outside.
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global NODES COEFF FourierInfo 
% apply ifft to the 2nd variable of child 1, and first variable of child 2.
temp=COEFF(:,:,child1);
phil=fftshift(ifft(fftshift(temp')));
phir=fftshift(ifft(fftshift(COEFF(:,:,child2))));

k=-FourierInfo.num:FourierInfo.num-1;
weight=exp(-k.^2*pi^2/4/(FourierInfo.LPeriod)^2);

% find the parent index.
inode=NODES(4,child1);
% the output is a matrix.
for m=-FourierInfo.num:FourierInfo.num-1,
    phinew1=(phil(:,m+1+FourierInfo.num)*ones([1,2*FourierInfo.num])).*phir;
    coefm=weight*fftshift(fft(fftshift(phinew1)))/2/FourierInfo.num;
    COEFF(m+FourierInfo.num+1,:,inode)= coefm;
    % below is the unvectorized version.
    % for j=-FourierInfo.num:FourierInfo.num-1,
      %   phinew=phil(:,m+1+FourierInfo.num).*phir(:,j+1+FourierInfo.num);
      %   coefmj=weight*fftshift(fft(fftshift(phinew)))/2/FourierInfo.num;
      %   COEFF(m+FourierInfo.num+1,j+FourierInfo.num+1,inode)= coefmj;
    % end 
end

end