%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% create the tree structure.
% 
% input:
%   ntotal = 2^L: the dimension of the problem.
%   NODES(:,j): stored information for each node.
%       1 - level, 
%       2 - id within level, starting from 1. For leaf node, this tells
%           which x-variable to integrate.
%       3 - =1,2,3,or 4, boundary or interior node.
%           0: root node.
%           1: left boundary, 2: right boundary
%           3: left interior, 4: right interior
%       4 - parent, 
%       5~6 - children, 
%       7~8 - the starting and ending points of the interval.
%       9~10 - the starting and ending location index of points in the interval.
%       11 - not used.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear global; clear variables; close all; clc; format long e;
addpath('./../nufft2d-1.3.2/')

% global variables.
%   NODES: tree structure
%   COEFF: the function coefficients for each node point, two variables.
%           updated in upward pass.
%   TRANSCoef(1:11,#ofNodes): the children-parent coefficient relations,the length
%        of the fundamental period, and the number of Fourier terms for each variable.
%        w1=c1*w1p+c2*w2p+c3*tnew;
%        w2=c4*w1p+c5*w2p+c6*tnew;
%           w1p,w2p are the parent's 2 variables. 
%        TRANSCoef(1:6,inode): store c1-c6. 
%        TRANSCoef(7:8,inode): the interval length of w1 and w2. Outside,
%           the integrand can be considered as 0.
%        TRANSCoef(9-11,inode): the number of Fourier Coefficients
%           for w1, w2 of current node, and t-variable used to decouple
%           parent node.
%   ABvec(1:2,ntotal): the integration bound for each x-variable. 
%   ZLoc(ntotal): the location of the z-variables.
%
%   Amatrix: the original and updated matrix for each level, for debugging
%       purposes only.

global ZLoc NODES COEFF ABvec TRANSCoef Amatrix

% test 2^ntotal_test cases.
ntotal_test(1) = 9; % number of levels
ntotal = 2^ntotal_test(1) % number of node points.
nnodes = 2*ntotal-1;
TRANSCoef(1:11,1:nnodes)=zeros([11,nnodes]);
TRANSCoef(9:11,1:nnodes)=32*ones([3,nnodes]); % Number of Fourier terms 
% This needs to be changed later for each node.
        
% construct the observation points.
rng(1);ztemp=rand([ntotal,1]);ZLoc=sort(ztemp); % the locations of the Z-variables.
% construct the matrix explictly for debugging purposes.
Amatrix=zeros([ntotal,ntotal]);
for m=1:ntotal,
    Amatrix(m,:)=exp(-abs(ZLoc(m)-ZLoc)); 
    % A is s.p.d. Note: this storage is not required. This is for 
    %   debuging purpose only.
end

NODES = zeros(11,2*ntotal-1); % the tree structure.

ABvec=zeros(2,ntotal); 
ABvec(1,:)=-ones(1,ntotal); ABvec(2,:)=ones(1,ntotal);
ABvec(2,1)=0.5;ABvec(2,2)=2; % no symmetric example, two points, compare with Mathematica to show 16 digits.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Construct the uniform binary tree.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% root node at level 0.
inode_now=1; % current node.
levelnodestart=zeros([ntotal_test(1)+1,1]); levelnodeend=levelnodestart;
levelnodestart(1)=1; % the starting node at level 0.
levelnodeend(1)=1; % the end node at level 0.
NODES(1,1) = 0; NODES(2,1) = 1; NODES(3,1) = 0; % root node
NODES(4,1) = 0; % no parent
NODES(5,1) = 0; NODES(6,1) = 0; % currently no children.
NODES(7,1) = 0; NODES(8,1) = 1; % all the points are located between [0,1].
NODES(9,1) = 1; NODES(10,1)=ntotal; % starting and ending nodes in the root interval.

% generate the required NODES info for all levels.
for ilevel=1:ntotal_test(1)
    levelnodestart(ilevel+1)=inode_now+1; % the start node in this level.
    for iparent=levelnodestart(ilevel):levelnodeend(ilevel)
        % generate the two child nodes.
        inode_now=inode_now+1; ichild1=inode_now;
        inode_now=inode_now+1; ichild2=inode_now;
        
        NODES(5,iparent)=ichild1; NODES(6,iparent)=ichild2; %parent's children.
        
        NODES(1,ichild1)=ilevel; NODES(1,ichild2)=ilevel;
        NODES(2,ichild1)=ichild1-levelnodestart(ilevel+1)+1; 
        NODES(2,ichild2)=ichild2-levelnodestart(ilevel+1)+1;
        
        if NODES(3,iparent)==0
            NODES(3,ichild1)=1; NODES(3,ichild2)=2;
        elseif NODES(3,iparent)==1
            NODES(3,ichild1)=1; NODES(3,ichild2)=4;
        elseif NODES(3,iparent)==2
            NODES(3,ichild1)=3; NODES(3,ichild2)=2;
        else
            NODES(3,ichild1)=3; NODES(3,ichild2)=4;
        end
        
        NODES(4,ichild1)=iparent; NODES(4,ichild2)=iparent;
        NODES(5,ichild1)=0; NODES(6,ichild1)=0;
        NODES(5,ichild2)=0; NODES(6,ichild2)=0;
        NODES(9,ichild1)=NODES(9,iparent); 
        NODES(10,ichild2)=NODES(10,iparent);
        midnode=(NODES(9,iparent)+NODES(10,iparent)-1)/2;
        NODES(10,ichild1)=midnode; NODES(9,ichild2)=midnode+1;
        
        NODES(7,ichild1)=NODES(7,iparent);
        NODES(8,ichild2)=NODES(8,iparent);
        midpoint=(ZLoc(midnode)+ZLoc(midnode+1))/2;
        NODES(8,ichild1)=midpoint;
        NODES(7,ichild2)=midpoint;
    end
    levelnodeend(ilevel+1)=inode_now;
end
%finished tree node generation.

tic
% perform the downward pass recursively to find the relations between
%   parent and children nodes.
compexp_downward(1);

% perform the upward pass recursively.
compexp_upward(1);
toc

COEFF(1,1,1)   % the computed integral.
